# The MIT License
#
# Copyright (c) 2008 Bob Farrell
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pdb
import bpython


class BPdb(pdb.Pdb):
    """PDB with BPython support."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.prompt = "(BPdb) "
        self.intro = 'Use "B" to enter bpython, Ctrl-d to exit it.'

    def postloop(self) -> None:
        # We only want to show the intro message once.
        self.intro = None
        super().postloop()

    # cmd.Cmd commands

    def do_Bpython(self, arg: str) -> None:
        locals_ = self.curframe.f_globals.copy()
        locals_.update(self.curframe.f_locals)
        bpython.embed(locals_, ["-i"])

    def help_Bpython(self) -> None:
        print("B(python)")
        print("")
        print(
            "Invoke the bpython interpreter for this stack frame. To exit "
            "bpython and return to a standard pdb press Ctrl-d"
        )

    # shortcuts
    do_B = do_Bpython
    help_B = help_Bpython



import errno
import os
import pty
import re
import select
import subprocess
import sys
import tempfile
import unittest

from textwrap import dedent
from bpython import args
from bpython.config import getpreferredencoding
from bpython.test import FixLanguageTestCase as TestCase


def run_with_tty(command):
    # based on https://stackoverflow.com/questions/52954248/capture-output-as-a-tty-in-python
    master_stdout, slave_stdout = pty.openpty()
    master_stderr, slave_stderr = pty.openpty()
    master_stdin, slave_stdin = pty.openpty()

    p = subprocess.Popen(
        command,
        stdout=slave_stdout,
        stderr=slave_stderr,
        stdin=slave_stdin,
        close_fds=True,
    )
    for fd in (slave_stdout, slave_stderr, slave_stdin):
        os.close(fd)

    readable = [master_stdout, master_stderr]
    result = {master_stdout: b"", master_stderr: b""}
    try:
        while readable:
            ready, _, _ = select.select(readable, [], [], 1)
            for fd in ready:
                try:
                    data = os.read(fd, 512)
                except OSError as e:
                    if e.errno != errno.EIO:
                        raise
                    # EIO means EOF on some systems
                    readable.remove(fd)
                else:
                    if not data:  # EOF
                        readable.remove(fd)
                    result[fd] += data
    finally:
        for fd in (master_stdout, master_stderr, master_stdin):
            os.close(fd)
        if p.poll() is None:
            p.kill()
        p.wait()

    if p.returncode:
        raise RuntimeError(f"Subprocess exited with {p.returncode}")

    return (
        result[master_stdout].decode(getpreferredencoding()),
        result[master_stderr].decode(getpreferredencoding()),
    )


class TestExecArgs(unittest.TestCase):
    def test_exec_dunder_file(self):
        with tempfile.NamedTemporaryFile(mode="w") as f:
            f.write(
                dedent(
                    """\
                import sys
                sys.stderr.write(__file__)
                sys.stderr.flush()"""
                )
            )
            f.flush()
            _, stderr = run_with_tty(
                [sys.executable] + ["-m", "bpython.curtsies", f.name]
            )
            self.assertEqual(stderr.strip(), f.name)

    def test_exec_nonascii_file(self):
        with tempfile.NamedTemporaryFile(mode="w") as f:
            f.write(
                dedent(
                    """\
                # coding: utf-8
                "你好 # nonascii"
                """
                )
            )
            f.flush()
            _, stderr = run_with_tty(
                [sys.executable, "-m", "bpython.curtsies", f.name],
            )
            self.assertEqual(len(stderr), 0)

    def test_exec_nonascii_file_linenums(self):
        with tempfile.NamedTemporaryFile(mode="w") as f:
            f.write(
                dedent(
                    """\
                1/0
                """
                )
            )
            f.flush()
            _, stderr = run_with_tty(
                [sys.executable, "-m", "bpython.curtsies", f.name],
            )
            self.assertIn("line 1", clean_colors(stderr))


def clean_colors(s):
    return re.sub(r"\x1b[^m]*m", "", s)


class TestParse(TestCase):
    def test_version(self):
        with self.assertRaises(SystemExit):
            args.parse(["--version"])



import os
from typing import cast

from bpython.test import FixLanguageTestCase as TestCase, TEST_CONFIG
from bpython.curtsiesfrontend import repl as curtsiesrepl
from bpython import config

from curtsies.window import CursorAwareWindow


def setup_config(conf):
    config_struct = config.Config(TEST_CONFIG)
    for key, value in conf.items():
        if not hasattr(config_struct, key):
            raise ValueError(f"{key!r} is not a valid config attribute")
        setattr(config_struct, key, value)
    return config_struct


def create_repl(brackets_enabled=False, **kwargs):
    config = setup_config(
        {"editor": "true", "brackets_completion": brackets_enabled}
    )
    repl = curtsiesrepl.BaseRepl(
        config, cast(CursorAwareWindow, None), **kwargs
    )
    os.environ["PAGER"] = "true"
    os.environ.pop("PYTHONSTARTUP", None)
    repl.width = 50
    repl.height = 20
    return repl


class TestBracketCompletionEnabled(TestCase):
    def setUp(self):
        self.repl = create_repl(brackets_enabled=True)

    def process_multiple_events(self, event_list):
        for event in event_list:
            self.repl.process_event(event)

    def test_start_line(self):
        self.repl.process_event("(")
        self.assertEqual(self.repl._current_line, "()")
        self.assertEqual(self.repl._cursor_offset, 1)

    def test_nested_brackets(self):
        self.process_multiple_events(["(", "[", "{"])
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 3)

    def test_quotes(self):
        self.process_multiple_events(["(", "'", "x", "<TAB>", ","])
        self.process_multiple_events(["[", '"', "y", "<TAB>", "<TAB>", "<TAB>"])
        self.assertEqual(self.repl._current_line, """('x',["y"])""")
        self.assertEqual(self.repl._cursor_offset, 11)

    def test_bracket_overwrite_closing_char(self):
        self.process_multiple_events(["(", "[", "{"])
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 3)
        self.process_multiple_events(["}", "]", ")"])
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 6)

    def test_brackets_move_cursor_on_tab(self):
        self.process_multiple_events(["(", "[", "{"])
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 3)
        self.repl.process_event("<TAB>")
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 4)
        self.repl.process_event("<TAB>")
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 5)
        self.repl.process_event("<TAB>")
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 6)

    def test_brackets_non_whitespace_following_char(self):
        self.repl.current_line = "s = s.connect('localhost', 8080)"
        self.repl.cursor_offset = 14
        self.repl.process_event("(")
        self.assertEqual(
            self.repl._current_line, "s = s.connect(('localhost', 8080)"
        )
        self.assertEqual(self.repl._cursor_offset, 15)

    def test_brackets_deletion_on_backspace(self):
        self.repl.current_line = "def foo()"
        self.repl.cursor_offset = 8
        self.repl.process_event("<BACKSPACE>")
        self.assertEqual(self.repl._current_line, "def foo")
        self.assertEqual(self.repl.cursor_offset, 7)

    def test_brackets_deletion_on_backspace_nested(self):
        self.repl.current_line = '([{""}])'
        self.repl.cursor_offset = 4
        self.process_multiple_events(
            ["<BACKSPACE>", "<BACKSPACE>", "<BACKSPACE>"]
        )
        self.assertEqual(self.repl._current_line, "()")
        self.assertEqual(self.repl.cursor_offset, 1)


class TestBracketCompletionDisabled(TestCase):
    def setUp(self):
        self.repl = create_repl(brackets_enabled=False)

    def process_multiple_events(self, event_list):
        for event in event_list:
            self.repl.process_event(event)

    def test_start_line(self):
        self.repl.process_event("(")
        self.assertEqual(self.repl._current_line, "(")
        self.assertEqual(self.repl._cursor_offset, 1)

    def test_nested_brackets(self):
        self.process_multiple_events(["(", "[", "{"])
        self.assertEqual(self.repl._current_line, "([{")
        self.assertEqual(self.repl._cursor_offset, 3)

    def test_bracket_overwrite_closing_char(self):
        self.process_multiple_events(["(", "[", "{"])
        self.assertEqual(self.repl._current_line, """([{""")
        self.assertEqual(self.repl._cursor_offset, 3)
        self.process_multiple_events(["}", "]", ")"])
        self.assertEqual(self.repl._current_line, """([{}])""")
        self.assertEqual(self.repl._cursor_offset, 6)

    def test_brackets_move_cursor_on_tab(self):
        self.process_multiple_events(["(", "[", "{"])
        self.assertEqual(self.repl._current_line, """([{""")
        self.assertEqual(self.repl._cursor_offset, 3)
        self.repl.process_event("<TAB>")
        self.assertEqual(self.repl._current_line, """([{""")
        self.assertEqual(self.repl._cursor_offset, 3)

    def test_brackets_deletion_on_backspace(self):
        self.repl.current_line = "def foo()"
        self.repl.cursor_offset = 8
        self.repl.process_event("<BACKSPACE>")
        self.assertEqual(self.repl._current_line, "def foo")
        self.assertEqual(self.repl.cursor_offset, 7)

    def test_brackets_deletion_on_backspace_nested(self):
        self.repl.current_line = '([{""}])'
        self.repl.cursor_offset = 4
        self.process_multiple_events(
            ["<BACKSPACE>", "<BACKSPACE>", "<BACKSPACE>"]
        )
        self.assertEqual(self.repl._current_line, "()")
        self.assertEqual(self.repl.cursor_offset, 1)



"""Tools for preparing code to be run in the REPL (removing blank lines,
etc)"""

from codeop import CommandCompiler
from re import Match
from itertools import tee, islice, chain

from ..lazyre import LazyReCompile

# TODO specifically catch IndentationErrors instead of any syntax errors

indent_empty_lines_re = LazyReCompile(r"\s*")
tabs_to_spaces_re = LazyReCompile(r"^\t+")


def indent_empty_lines(s: str, compiler: CommandCompiler) -> str:
    """Indents blank lines that would otherwise cause early compilation

    Only really works if starting on a new line"""
    initial_lines = s.split("\n")
    ends_with_newline = False
    if initial_lines and not initial_lines[-1]:
        ends_with_newline = True
        initial_lines.pop()
    result_lines = []

    prevs, lines, nexts = tee(initial_lines, 3)
    prevs = chain(("",), prevs)
    nexts = chain(islice(nexts, 1, None), ("",))

    for p_line, line, n_line in zip(prevs, lines, nexts):
        if len(line) == 0:
            # "\s*" always matches
            p_indent = indent_empty_lines_re.match(p_line).group()  # type: ignore
            n_indent = indent_empty_lines_re.match(n_line).group()  # type: ignore
            result_lines.append(min([p_indent, n_indent], key=len) + line)
        else:
            result_lines.append(line)

    return "\n".join(result_lines) + ("\n" if ends_with_newline else "")


def leading_tabs_to_spaces(s: str) -> str:
    def tab_to_space(m: Match[str]) -> str:
        return len(m.group()) * 4 * " "

    return "\n".join(
        tabs_to_spaces_re.sub(tab_to_space, line) for line in s.split("\n")
    )


def preprocess(s: str, compiler: CommandCompiler) -> str:
    return indent_empty_lines(leading_tabs_to_spaces(s), compiler)



import linecache
from typing import Any


class BPythonLinecache(dict):
    """Replaces the cache dict in the standard-library linecache module,
    to also remember (in an unerasable way) bpython console input."""

    def __init__(
        self,
        bpython_history: None | (list[tuple[int, None, list[str], str]]) = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.bpython_history = bpython_history or []

    def is_bpython_filename(self, fname: Any) -> bool:
        return isinstance(fname, str) and fname.startswith("<bpython-input-")

    def get_bpython_history(self, key: str) -> tuple[int, None, list[str], str]:
        """Given a filename provided by remember_bpython_input,
        returns the associated source string."""
        try:
            idx = int(key.split("-")[2][:-1])
            return self.bpython_history[idx]
        except (IndexError, ValueError):
            raise KeyError

    def remember_bpython_input(self, source: str) -> str:
        """Remembers a string of source code, and returns
        a fake filename to use to retrieve it later."""
        filename = f"<bpython-input-{len(self.bpython_history)}>"
        self.bpython_history.append(
            (len(source), None, source.splitlines(True), filename)
        )
        return filename

    def __getitem__(self, key: Any) -> Any:
        if self.is_bpython_filename(key):
            return self.get_bpython_history(key)
        return super().__getitem__(key)

    def __contains__(self, key: Any) -> bool:
        if self.is_bpython_filename(key):
            try:
                self.get_bpython_history(key)
                return True
            except KeyError:
                return False
        return super().__contains__(key)

    def __delitem__(self, key: Any) -> None:
        if not self.is_bpython_filename(key):
            super().__delitem__(key)


def _bpython_clear_linecache() -> None:
    if isinstance(linecache.cache, BPythonLinecache):
        bpython_history = linecache.cache.bpython_history
    else:
        bpython_history = None
    linecache.cache = BPythonLinecache(bpython_history)


# Monkey-patch the linecache module so that we are able
# to hold our command history there and have it persist
linecache.cache = BPythonLinecache(None, linecache.cache)  # type: ignore
linecache.clearcache = _bpython_clear_linecache


def filename_for_console_input(code_string: str) -> str:
    """Remembers a string of source code, and returns
    a fake filename to use to retrieve it later."""
    if isinstance(linecache.cache, BPythonLinecache):
        return linecache.cache.remember_bpython_input(code_string)
    else:
        # If someone else has patched linecache.cache, better for code to
        # simply be unavailable to inspect.getsource() than to raise
        # an exception.
        return "<input>"



import sys
import unittest

from curtsies.fmtfuncs import bold, green, magenta, cyan, red, plain

from bpython.curtsiesfrontend import interpreter

pypy = "PyPy" in sys.version


class Interpreter(interpreter.Interp):
    def __init__(self):
        super().__init__()
        self.a = []
        self.write = self.a.append


class TestInterpreter(unittest.TestCase):
    def test_syntaxerror(self):
        i = Interpreter()

        i.runsource("1.1.1.1")

        if (3, 10, 1) <= sys.version_info[:3]:
            expected = (
                "  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + "\n    1.1.1.1\n       ^^\n"
                + bold(red("SyntaxError"))
                + ": "
                + cyan("invalid syntax")
                + "\n"
            )
        elif (3, 10) <= sys.version_info[:2]:
            expected = (
                "  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + "\n    1.1.1.1\n    ^^^^^\n"
                + bold(red("SyntaxError"))
                + ": "
                + cyan("invalid syntax. Perhaps you forgot a comma?")
                + "\n"
            )
        elif (3, 8) <= sys.version_info[:2]:
            expected = (
                "  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + "\n    1.1.1.1\n       ^\n"
                + bold(red("SyntaxError"))
                + ": "
                + cyan("invalid syntax")
                + "\n"
            )
        elif pypy:
            expected = (
                "  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + "\n    1.1.1.1\n       ^\n"
                + bold(red("SyntaxError"))
                + ": "
                + cyan("invalid syntax")
                + "\n"
            )
        else:
            expected = (
                "  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + "\n    1.1.1.1\n        ^\n"
                + bold(red("SyntaxError"))
                + ": "
                + cyan("invalid syntax")
                + "\n"
            )

        a = i.a
        self.assertMultiLineEqual(str(plain("").join(a)), str(expected))
        self.assertEqual(plain("").join(a), expected)

    def test_traceback(self):
        i = Interpreter()

        def f():
            return 1 / 0

        def gfunc():
            return f()

        i.runsource("gfunc()")

        global_not_found = "name 'gfunc' is not defined"

        if (3, 13) <= sys.version_info[:2]:
            expected = (
                "Traceback (most recent call last):\n  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + ", in "
                + cyan("<module>")
                + "\n    gfunc()"
                + "\n    ^^^^^\n"
                + bold(red("NameError"))
                + ": "
                + cyan(global_not_found)
                + "\n"
            )
        elif (3, 11) <= sys.version_info[:2]:
            expected = (
                "Traceback (most recent call last):\n  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + ", in "
                + cyan("<module>")
                + "\n    gfunc()"
                + "\n     ^^^^^\n"
                + bold(red("NameError"))
                + ": "
                + cyan(global_not_found)
                + "\n"
            )
        else:
            expected = (
                "Traceback (most recent call last):\n  File "
                + green('"<input>"')
                + ", line "
                + bold(magenta("1"))
                + ", in "
                + cyan("<module>")
                + "\n    gfunc()\n"
                + bold(red("NameError"))
                + ": "
                + cyan(global_not_found)
                + "\n"
            )

        a = i.a
        self.assertMultiLineEqual(str(expected), str(plain("").join(a)))
        self.assertEqual(expected, plain("").join(a))

    def test_getsource_works_on_interactively_defined_functions(self):
        source = "def foo(x):\n    return x + 1\n"
        i = interpreter.Interp()
        i.runsource(source)
        import inspect

        inspected_source = inspect.getsource(i.locals["foo"])
        self.assertEqual(inspected_source, source)



import collections
import inspect
import os
import socket
import sys
import tempfile
import unittest

from typing import List, Tuple
from itertools import islice
from pathlib import Path
from unittest import mock

from bpython import config, repl, autocomplete
from bpython.line import LinePart
from bpython.test import (
    MagicIterMock,
    FixLanguageTestCase as TestCase,
    TEST_CONFIG,
)


pypy = "PyPy" in sys.version


def setup_config(conf):
    config_struct = config.Config(TEST_CONFIG)
    if conf is not None and "autocomplete_mode" in conf:
        config_struct.autocomplete_mode = conf["autocomplete_mode"]
    return config_struct


class FakeHistory(repl.History):
    def __init__(self):
        pass

    def reset(self):
        pass


class FakeRepl(repl.Repl):
    def __init__(self, conf=None):
        super().__init__(repl.Interpreter(), setup_config(conf))
        self._current_line = ""
        self._cursor_offset = 0

    def _get_current_line(self) -> str:
        return self._current_line

    def _set_current_line(self, val: str) -> None:
        self._current_line = val

    def _get_cursor_offset(self) -> int:
        return self._cursor_offset

    def _set_cursor_offset(self, val: int) -> None:
        self._cursor_offset = val

    def getstdout(self) -> str:
        raise NotImplementedError

    def reprint_line(
        self, lineno: int, tokens: list[tuple[repl._TokenType, str]]
    ) -> None:
        raise NotImplementedError

    def reevaluate(self):
        raise NotImplementedError


class TestMatchesIterator(unittest.TestCase):
    def setUp(self):
        self.matches = ["bobby", "bobbies", "bobberina"]
        self.matches_iterator = repl.MatchesIterator()
        self.matches_iterator.current_word = "bob"
        self.matches_iterator.orig_line = "bob"
        self.matches_iterator.orig_cursor_offset = len("bob")
        self.matches_iterator.matches = self.matches

    def test_next(self):
        self.assertEqual(next(self.matches_iterator), self.matches[0])

        for x in range(len(self.matches) - 1):
            next(self.matches_iterator)

        self.assertEqual(next(self.matches_iterator), self.matches[0])
        self.assertEqual(next(self.matches_iterator), self.matches[1])
        self.assertNotEqual(next(self.matches_iterator), self.matches[1])

    def test_previous(self):
        self.assertEqual(self.matches_iterator.previous(), self.matches[2])

        for x in range(len(self.matches) - 1):
            self.matches_iterator.previous()

        self.assertNotEqual(self.matches_iterator.previous(), self.matches[0])
        self.assertEqual(self.matches_iterator.previous(), self.matches[1])
        self.assertEqual(self.matches_iterator.previous(), self.matches[0])

    def test_nonzero(self):
        """self.matches_iterator should be False at start,
        then True once we active a match.
        """
        self.assertFalse(self.matches_iterator)
        next(self.matches_iterator)
        self.assertTrue(self.matches_iterator)

    def test_iter(self):
        slice = islice(self.matches_iterator, 0, 9)
        self.assertEqual(list(slice), self.matches * 3)

    def test_current(self):
        with self.assertRaises(ValueError):
            self.matches_iterator.current()
        next(self.matches_iterator)
        self.assertEqual(self.matches_iterator.current(), self.matches[0])

    def test_update(self):
        slice = islice(self.matches_iterator, 0, 3)
        self.assertEqual(list(slice), self.matches)

        newmatches = ["string", "str", "set"]
        completer = mock.Mock()
        completer.locate.return_value = LinePart(0, 1, "s")
        self.matches_iterator.update(1, "s", newmatches, completer)

        newslice = islice(newmatches, 0, 3)
        self.assertNotEqual(list(slice), self.matches)
        self.assertEqual(list(newslice), newmatches)

    def test_cur_line(self):
        completer = mock.Mock()
        completer.locate.return_value = LinePart(
            0,
            self.matches_iterator.orig_cursor_offset,
            self.matches_iterator.orig_line,
        )
        self.matches_iterator.completer = completer

        with self.assertRaises(ValueError):
            self.matches_iterator.cur_line()

        self.assertEqual(next(self.matches_iterator), self.matches[0])
        self.assertEqual(
            self.matches_iterator.cur_line(),
            (len(self.matches[0]), self.matches[0]),
        )

    def test_is_cseq(self):
        self.assertTrue(self.matches_iterator.is_cseq())


class TestArgspec(unittest.TestCase):
    def setUp(self):
        self.repl = FakeRepl()
        self.repl.push("def spam(a, b, c):\n", False)
        self.repl.push("    pass\n", False)
        self.repl.push("\n", False)
        self.repl.push("class Spam(object):\n", False)
        self.repl.push("    def spam(self, a, b, c):\n", False)
        self.repl.push("        pass\n", False)
        self.repl.push("\n", False)
        self.repl.push("class SpammitySpam(object):\n", False)
        self.repl.push("    def __init__(self, a, b, c):\n", False)
        self.repl.push("        pass\n", False)
        self.repl.push("\n", False)
        self.repl.push("class WonderfulSpam(object):\n", False)
        self.repl.push("    def __new__(self, a, b, c):\n", False)
        self.repl.push("        pass\n", False)
        self.repl.push("\n", False)
        self.repl.push("o = Spam()\n", False)
        self.repl.push("\n", False)

    def set_input_line(self, line):
        """Set current input line of the test REPL."""
        self.repl.current_line = line
        self.repl.cursor_offset = len(line)

    def test_func_name(self):
        for line, expected_name in [
            ("spam(", "spam"),
            # map pydoc has no signature in pypy
            ("spam(any([]", "any") if pypy else ("spam(map([]", "map"),
            ("spam((), ", "spam"),
        ]:
            self.set_input_line(line)
            self.assertTrue(self.repl.get_args())
            self.assertEqual(self.repl.current_func.__name__, expected_name)

    def test_func_name_method_issue_479(self):
        for line, expected_name in [
            ("o.spam(", "spam"),
            # map pydoc has no signature in pypy
            ("o.spam(any([]", "any") if pypy else ("o.spam(map([]", "map"),
            ("o.spam((), ", "spam"),
        ]:
            self.set_input_line(line)
            self.assertTrue(self.repl.get_args())
            self.assertEqual(self.repl.current_func.__name__, expected_name)

    def test_syntax_error_parens(self):
        for line in ["spam(]", "spam([)", "spam())"]:
            self.set_input_line(line)
            # Should not explode
            self.repl.get_args()

    def test_kw_arg_position(self):
        self.set_input_line("spam(a=0")
        self.assertTrue(self.repl.get_args())
        self.assertEqual(self.repl.arg_pos, "a")

        self.set_input_line("spam(1, b=1")
        self.assertTrue(self.repl.get_args())
        self.assertEqual(self.repl.arg_pos, "b")

        self.set_input_line("spam(1, c=2")
        self.assertTrue(self.repl.get_args())
        self.assertEqual(self.repl.arg_pos, "c")

    def test_lambda_position(self):
        self.set_input_line("spam(lambda a, b: 1, ")
        self.assertTrue(self.repl.get_args())
        self.assertTrue(self.repl.funcprops)
        # Argument position
        self.assertEqual(self.repl.arg_pos, 1)

    @unittest.skipIf(pypy, "range pydoc has no signature in pypy")
    def test_issue127(self):
        self.set_input_line("x=range(")
        self.assertTrue(self.repl.get_args())
        self.assertEqual(self.repl.current_func.__name__, "range")

        self.set_input_line("{x:range(")
        self.assertTrue(self.repl.get_args())
        self.assertEqual(self.repl.current_func.__name__, "range")

        self.set_input_line("foo(1, 2, x,range(")
        self.assertEqual(self.repl.current_func.__name__, "range")

        self.set_input_line("(x,range(")
        self.assertEqual(self.repl.current_func.__name__, "range")

    def test_nonexistent_name(self):
        self.set_input_line("spamspamspam(")
        self.assertFalse(self.repl.get_args())

    def test_issue572(self):
        self.set_input_line("SpammitySpam(")
        self.assertTrue(self.repl.get_args())

        self.set_input_line("WonderfulSpam(")
        self.assertTrue(self.repl.get_args())

    @unittest.skipIf(pypy, "pypy pydoc doesn't have this")
    def test_issue583(self):
        self.repl = FakeRepl()
        self.repl.push("a = 1.2\n", False)
        self.set_input_line("a.is_integer(")
        self.repl.set_docstring()
        self.assertIsNot(self.repl.docstring, None)

    def test_methods_of_expressions(self):
        self.set_input_line("'a'.capitalize(")
        self.assertTrue(self.repl.get_args())

        self.set_input_line("(1 + 1.1).as_integer_ratio(")
        self.assertTrue(self.repl.get_args())


class TestArgspecInternal(unittest.TestCase):
    def test_function_expressions(self):
        te = self.assertTupleEqual
        fa = lambda line: repl.Repl._funcname_and_argnum(line)
        for line, (func, argnum) in [
            ("spam(", ("spam", 0)),
            ("spam((), ", ("spam", 1)),
            ("spam.eggs((), ", ("spam.eggs", 1)),
            ("spam[abc].eggs((), ", ("spam[abc].eggs", 1)),
            ("spam[0].eggs((), ", ("spam[0].eggs", 1)),
            ("spam[a + b]eggs((), ", ("spam[a + b]eggs", 1)),
            ("spam().eggs((), ", ("spam().eggs", 1)),
            ("spam(1, 2).eggs((), ", ("spam(1, 2).eggs", 1)),
            ("spam(1, f(1)).eggs((), ", ("spam(1, f(1)).eggs", 1)),
            ("[0].eggs((), ", ("[0].eggs", 1)),
            ("[0][0]((), {}).eggs((), ", ("[0][0]((), {}).eggs", 1)),
            ("a + spam[0].eggs((), ", ("spam[0].eggs", 1)),
            ("spam(", ("spam", 0)),
            ("spam(map([]", ("map", 0)),
            ("spam((), ", ("spam", 1)),
        ]:
            te(fa(line), (func, argnum))


class TestGetSource(unittest.TestCase):
    def setUp(self):
        self.repl = FakeRepl()

    def set_input_line(self, line):
        """Set current input line of the test REPL."""
        self.repl.current_line = line
        self.repl.cursor_offset = len(line)

    def assert_get_source_error_for_current_function(self, func, msg):
        self.repl.current_func = func
        with self.assertRaises(repl.SourceNotFound):
            self.repl.get_source_of_current_name()
        try:
            self.repl.get_source_of_current_name()
        except repl.SourceNotFound as e:
            self.assertEqual(msg, e.args[0])
        else:
            self.fail("Should have raised SourceNotFound")

    def test_current_function(self):
        self.set_input_line("INPUTLINE")
        self.repl.current_func = inspect.getsource
        self.assertIn(
            "text of the source code", self.repl.get_source_of_current_name()
        )

        self.assert_get_source_error_for_current_function(
            [], "No source code found for INPUTLINE"
        )

        self.assert_get_source_error_for_current_function(
            list.pop, "No source code found for INPUTLINE"
        )

    @unittest.skipIf(pypy, "different errors for PyPy")
    def test_current_function_cpython(self):
        self.set_input_line("INPUTLINE")
        self.assert_get_source_error_for_current_function(
            collections.defaultdict.copy, "No source code found for INPUTLINE"
        )
        if sys.version_info[:2] >= (3, 13):
            self.assert_get_source_error_for_current_function(
                collections.defaultdict, "source code not available"
            )
        else:
            self.assert_get_source_error_for_current_function(
                collections.defaultdict, "could not find class definition"
            )

    def test_current_line(self):
        self.repl.interp.locals["a"] = socket.socket
        self.set_input_line("a")
        self.assertIn("dup(self)", self.repl.get_source_of_current_name())


# TODO add tests for various failures without using current function


class TestEditConfig(TestCase):
    def setUp(self):
        self.repl = FakeRepl()
        self.repl.interact.confirm = lambda msg: True
        self.repl.interact.notify = lambda msg: None
        self.repl.config.editor = "true"

    def test_create_config(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            config_path = Path(tmp_dir) / "newdir" / "config"
            self.repl.config.config_path = config_path
            self.repl.edit_config()
            self.assertTrue(config_path.exists())


class TestRepl(unittest.TestCase):
    def set_input_line(self, line):
        """Set current input line of the test REPL."""
        self.repl.current_line = line
        self.repl.cursor_offset = len(line)

    def setUp(self):
        self.repl = FakeRepl()

    def test_current_string(self):
        self.set_input_line('a = "2"')
        # TODO factor cpos out of repl.Repl
        self.repl.cpos = 0
        self.assertEqual(self.repl.current_string(), '"2"')

        self.set_input_line('a = "2" + 2')
        self.assertEqual(self.repl.current_string(), "")

    def test_push(self):
        self.repl = FakeRepl()
        self.repl.push("foobar = 2")
        self.assertEqual(self.repl.interp.locals["foobar"], 2)

    # COMPLETE TESTS
    # 1. Global tests
    def test_simple_global_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SIMPLE}
        )
        self.set_input_line("d")

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(
            self.repl.matches_iter.matches,
            ["def", "del", "delattr(", "dict(", "dir(", "divmod("],
        )

    def test_substring_global_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SUBSTRING}
        )
        self.set_input_line("time")

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(
            self.repl.matches_iter.matches, ["RuntimeError(", "RuntimeWarning("]
        )

    def test_fuzzy_global_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.FUZZY}
        )
        self.set_input_line("doc")

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(
            self.repl.matches_iter.matches,
            ["ChildProcessError(", "UnboundLocalError(", "__doc__"],
        )

    # 2. Attribute tests
    def test_simple_attribute_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SIMPLE}
        )
        self.set_input_line("Foo.b")

        code = "class Foo():\n\tdef bar(self):\n\t\tpass\n"
        for line in code.split("\n"):
            self.repl.push(line)

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(self.repl.matches_iter.matches, ["Foo.bar"])

    def test_substring_attribute_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SUBSTRING}
        )
        self.set_input_line("Foo.az")

        code = "class Foo():\n\tdef baz(self):\n\t\tpass\n"
        for line in code.split("\n"):
            self.repl.push(line)

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(self.repl.matches_iter.matches, ["Foo.baz"])

    def test_fuzzy_attribute_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.FUZZY}
        )
        self.set_input_line("Foo.br")

        code = "class Foo():\n\tdef bar(self):\n\t\tpass\n"
        for line in code.split("\n"):
            self.repl.push(line)

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(self.repl.matches_iter.matches, ["Foo.bar"])

    # 3. Edge cases
    def test_updating_namespace_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SIMPLE}
        )
        self.set_input_line("foo")
        self.repl.push("foobar = 2")

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(self.repl.matches_iter.matches, ["foobar"])

    def test_file_should_not_appear_in_complete(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SIMPLE}
        )
        self.set_input_line("_")
        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertNotIn("__file__", self.repl.matches_iter.matches)

    # 4. Parameter names
    def test_paremeter_name_completion(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SIMPLE}
        )
        self.set_input_line("foo(ab")

        code = "def foo(abc=1, abd=2, xyz=3):\n\tpass\n"
        for line in code.split("\n"):
            self.repl.push(line)

        self.assertTrue(self.repl.complete())
        self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
        self.assertEqual(
            self.repl.matches_iter.matches, ["abc=", "abd=", "abs("]
        )

    def test_parameter_advanced_on_class(self):
        self.repl = FakeRepl(
            {"autocomplete_mode": autocomplete.AutocompleteModes.SIMPLE}
        )
        self.set_input_line("TestCls(app")

        code = """
        import inspect

        class TestCls:
            # A class with boring __init__ typing
            def __init__(self, *args, **kwargs):
                pass
            # But that uses super exotic typings recognized by inspect.signature
            __signature__ = inspect.Signature([
                inspect.Parameter("apple", inspect.Parameter.POSITIONAL_ONLY),
                inspect.Parameter("apple2", inspect.Parameter.KEYWORD_ONLY),
                inspect.Parameter("pinetree", inspect.Parameter.KEYWORD_ONLY),
            ])
        """
        code = [x[8:] for x in code.split("\n")]
        for line in code:
            self.repl.push(line)

        with mock.patch(
            "bpython.inspection.inspect.getsourcelines",
            return_value=(code, None),
        ):
            self.assertTrue(self.repl.complete())
            self.assertTrue(hasattr(self.repl.matches_iter, "matches"))
            self.assertEqual(
                self.repl.matches_iter.matches, ["apple2=", "apple="]
            )


if __name__ == "__main__":
    unittest.main()



# The MIT License
#
# Copyright (c) 2009-2015 the bpython authors.
# Copyright (c) 2015-2022 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# To gradually migrate to mypy we aren't setting these globally yet
# mypy: disallow_untyped_defs=True
# mypy: disallow_untyped_calls=True

import os
import sys
import locale
from configparser import ConfigParser
from itertools import chain
from pathlib import Path
from typing import Any, Dict
from collections.abc import MutableMapping, Mapping
from xdg import BaseDirectory

from .autocomplete import AutocompleteModes

default_completion = AutocompleteModes.SIMPLE
# All supported letters for colors for themes
#
# Instead of importing it from .curtsiesfrontend.parse, we define them here to
# avoid a potential import of fcntl on Windows.
COLOR_LETTERS = tuple("krgybmcwd")


class UnknownColorCode(Exception):
    def __init__(self, key: str, color: str) -> None:
        self.key = key
        self.color = color


def getpreferredencoding() -> str:
    """Get the user's preferred encoding."""
    return locale.getpreferredencoding() or sys.getdefaultencoding()


def can_encode(c: str) -> bool:
    try:
        c.encode(getpreferredencoding())
        return True
    except UnicodeEncodeError:
        return False


def supports_box_chars() -> bool:
    """Check if the encoding supports Unicode box characters."""
    return all(map(can_encode, "│─└┘┌┐"))


def get_config_home() -> Path:
    """Returns the base directory for bpython's configuration files."""
    return Path(BaseDirectory.xdg_config_home) / "bpython"


def default_config_path() -> Path:
    """Returns bpython's default configuration file path."""
    return get_config_home() / "config"


def default_editor() -> str:
    """Returns the default editor."""
    return os.environ.get("VISUAL", os.environ.get("EDITOR", "vi"))


def fill_config_with_default_values(
    config: ConfigParser, default_values: Mapping[str, Mapping[str, Any]]
) -> None:
    for section in default_values.keys():
        if not config.has_section(section):
            config.add_section(section)

        for opt, val in default_values[section].items():
            if not config.has_option(section, opt):
                config.set(section, opt, str(val))


class Config:
    default_colors = {
        "keyword": "y",
        "name": "c",
        "comment": "b",
        "string": "m",
        "error": "r",
        "number": "G",
        "operator": "Y",
        "punctuation": "y",
        "token": "C",
        "background": "d",
        "output": "w",
        "main": "c",
        "paren": "R",
        "prompt": "c",
        "prompt_more": "g",
        "right_arrow_suggestion": "K",
    }

    defaults: dict[str, dict[str, Any]] = {
        "general": {
            "arg_spec": True,
            "auto_display_list": True,
            "autocomplete_mode": default_completion,
            "color_scheme": "default",
            "complete_magic_methods": True,
            "dedent_after": 1,
            "default_autoreload": False,
            "editor": default_editor(),
            "flush_output": True,
            "import_completion_skiplist": ":".join(
                (
                    # version tracking
                    ".git",
                    ".svn",
                    ".hg"
                    # XDG
                    ".config",
                    ".local",
                    ".share",
                    # nodejs
                    "node_modules",
                    # PlayOnLinux
                    "PlayOnLinux's virtual drives",
                    # wine
                    "dosdevices",
                    # Python byte code cache
                    "__pycache__",
                )
            ),
            "highlight_show_source": True,
            "hist_duplicates": True,
            "hist_file": "~/.pythonhist",
            "hist_length": 1000,
            "paste_time": 0.02,
            "pastebin_confirm": True,
            "pastebin_expiry": "1week",
            "pastebin_helper": "",
            "pastebin_url": "https://bpaste.net",
            "save_append_py": False,
            "single_undo_time": 1.0,
            "syntax": True,
            "tab_length": 4,
            "unicode_box": True,
            "brackets_completion": False,
        },
        "keyboard": {
            "backspace": "C-h",
            "beginning_of_line": "C-a",
            "clear_line": "C-u",
            "clear_screen": "C-l",
            "clear_word": "C-w",
            "copy_clipboard": "F10",
            "cut_to_buffer": "C-k",
            "delete": "C-d",
            "down_one_line": "C-n",
            "edit_config": "F3",
            "edit_current_block": "C-x",
            "end_of_line": "C-e",
            "exit": "",
            "external_editor": "F7",
            "help": "F1",
            "incremental_search": "M-s",
            "last_output": "F9",
            "left": "C-b",
            "pastebin": "F8",
            "redo": "C-g",
            "reimport": "F6",
            "reverse_incremental_search": "M-r",
            "right": "C-f",
            "save": "C-s",
            "search": "C-o",
            "show_source": "F2",
            "suspend": "C-z",
            "toggle_file_watch": "F5",
            "transpose_chars": "C-t",
            "undo": "C-r",
            "up_one_line": "C-p",
            "yank_from_buffer": "C-y",
        },
        "cli": {
            "suggestion_width": 0.8,
            "trim_prompts": False,
        },
        "curtsies": {
            "list_above": False,
            "right_arrow_completion": True,
        },
    }

    def __init__(self, config_path: Path | None = None) -> None:
        """Loads .ini configuration file and stores its values."""

        config = ConfigParser()
        fill_config_with_default_values(config, self.defaults)
        try:
            if config_path is not None:
                config.read(config_path)
        except UnicodeDecodeError as e:
            sys.stderr.write(
                "Error: Unable to parse config file at '{}' due to an "
                "encoding issue ({}). Please make sure to fix the encoding "
                "of the file or remove it and then try again.\n".format(
                    config_path, e
                )
            )
            sys.exit(1)

        default_keys_to_commands = {
            value: key for (key, value) in self.defaults["keyboard"].items()
        }

        def get_key_no_doublebind(command: str) -> str:
            default_commands_to_keys = self.defaults["keyboard"]
            requested_key = config.get("keyboard", command)

            try:
                default_command = default_keys_to_commands[requested_key]
                if default_commands_to_keys[default_command] == config.get(
                    "keyboard", default_command
                ):
                    setattr(self, f"{default_command}_key", "")
            except KeyError:
                pass

            return requested_key

        self.config_path = (
            config_path.absolute() if config_path is not None else None
        )
        self.hist_file = Path(config.get("general", "hist_file")).expanduser()

        self.dedent_after = config.getint("general", "dedent_after")
        self.tab_length = config.getint("general", "tab_length")
        self.auto_display_list = config.getboolean(
            "general", "auto_display_list"
        )
        self.syntax = config.getboolean("general", "syntax")
        self.arg_spec = config.getboolean("general", "arg_spec")
        self.paste_time = config.getfloat("general", "paste_time")
        self.single_undo_time = config.getfloat("general", "single_undo_time")
        self.highlight_show_source = config.getboolean(
            "general", "highlight_show_source"
        )
        self.editor = config.get("general", "editor")
        self.hist_length = config.getint("general", "hist_length")
        self.hist_duplicates = config.getboolean("general", "hist_duplicates")
        self.flush_output = config.getboolean("general", "flush_output")
        self.default_autoreload = config.getboolean(
            "general", "default_autoreload"
        )
        self.import_completion_skiplist = config.get(
            "general", "import_completion_skiplist"
        ).split(":")

        self.pastebin_key = get_key_no_doublebind("pastebin")
        self.copy_clipboard_key = get_key_no_doublebind("copy_clipboard")
        self.save_key = get_key_no_doublebind("save")
        self.search_key = get_key_no_doublebind("search")
        self.show_source_key = get_key_no_doublebind("show_source")
        self.suspend_key = get_key_no_doublebind("suspend")
        self.toggle_file_watch_key = get_key_no_doublebind("toggle_file_watch")
        self.undo_key = get_key_no_doublebind("undo")
        self.redo_key = get_key_no_doublebind("redo")
        self.reimport_key = get_key_no_doublebind("reimport")
        self.reverse_incremental_search_key = get_key_no_doublebind(
            "reverse_incremental_search"
        )
        self.incremental_search_key = get_key_no_doublebind(
            "incremental_search"
        )
        self.up_one_line_key = get_key_no_doublebind("up_one_line")
        self.down_one_line_key = get_key_no_doublebind("down_one_line")
        self.cut_to_buffer_key = get_key_no_doublebind("cut_to_buffer")
        self.yank_from_buffer_key = get_key_no_doublebind("yank_from_buffer")
        self.clear_word_key = get_key_no_doublebind("clear_word")
        self.backspace_key = get_key_no_doublebind("backspace")
        self.clear_line_key = get_key_no_doublebind("clear_line")
        self.clear_screen_key = get_key_no_doublebind("clear_screen")
        self.delete_key = get_key_no_doublebind("delete")

        self.left_key = get_key_no_doublebind("left")
        self.right_key = get_key_no_doublebind("right")
        self.end_of_line_key = get_key_no_doublebind("end_of_line")
        self.beginning_of_line_key = get_key_no_doublebind("beginning_of_line")
        self.transpose_chars_key = get_key_no_doublebind("transpose_chars")
        self.exit_key = get_key_no_doublebind("exit")
        self.last_output_key = get_key_no_doublebind("last_output")
        self.edit_config_key = get_key_no_doublebind("edit_config")
        self.edit_current_block_key = get_key_no_doublebind(
            "edit_current_block"
        )
        self.external_editor_key = get_key_no_doublebind("external_editor")
        self.help_key = get_key_no_doublebind("help")

        self.pastebin_confirm = config.getboolean("general", "pastebin_confirm")
        self.pastebin_url = config.get("general", "pastebin_url")
        self.pastebin_expiry = config.get("general", "pastebin_expiry")
        self.pastebin_helper = config.get("general", "pastebin_helper")

        self.cli_suggestion_width = config.getfloat("cli", "suggestion_width")
        self.cli_trim_prompts = config.getboolean("cli", "trim_prompts")

        self.complete_magic_methods = config.getboolean(
            "general", "complete_magic_methods"
        )
        self.autocomplete_mode = (
            AutocompleteModes.from_string(
                config.get("general", "autocomplete_mode")
            )
            or default_completion
        )
        self.save_append_py = config.getboolean("general", "save_append_py")

        self.curtsies_list_above = config.getboolean("curtsies", "list_above")
        self.curtsies_right_arrow_completion = config.getboolean(
            "curtsies", "right_arrow_completion"
        )
        self.unicode_box = config.getboolean("general", "unicode_box")

        self.color_scheme = dict()
        color_scheme_name = config.get("general", "color_scheme")
        if color_scheme_name == "default":
            self.color_scheme.update(self.default_colors)
        else:
            path = get_config_home() / f"{color_scheme_name}.theme"
            try:
                load_theme(path, self.color_scheme, self.default_colors)
            except OSError:
                sys.stderr.write(
                    f"Could not load theme '{color_scheme_name}' from {path}.\n"
                )
                sys.exit(1)
            except UnknownColorCode as ucc:
                sys.stderr.write(
                    f"Theme '{color_scheme_name}' contains invalid color: {ucc.key} = {ucc.color}.\n"
                )
                sys.exit(1)

        # set box drawing characters
        (
            self.left_border,
            self.right_border,
            self.top_border,
            self.bottom_border,
            self.left_bottom_corner,
            self.right_bottom_corner,
            self.left_top_corner,
            self.right_top_corner,
        ) = (
            ("│", "│", "─", "─", "└", "┘", "┌", "┐")
            if self.unicode_box and supports_box_chars()
            else ("|", "|", "-", "-", "+", "+", "+", "+")
        )
        self.brackets_completion = config.getboolean(
            "general", "brackets_completion"
        )


def load_theme(
    path: Path,
    colors: MutableMapping[str, str],
    default_colors: Mapping[str, str],
) -> None:
    theme = ConfigParser()
    with open(path) as f:
        theme.read_file(f)
    for k, v in chain(theme.items("syntax"), theme.items("interface")):
        if theme.has_option("syntax", k):
            colors[k] = theme.get("syntax", k)
        else:
            colors[k] = theme.get("interface", k)
        if colors[k].lower() not in COLOR_LETTERS:
            raise UnknownColorCode(k, colors[k])

    # Check against default theme to see if all values are defined
    for k, v in default_colors.items():
        if k not in colors:
            colors[k] = v



#careful! Whitespace is very important in this file

class BlankLineBetweenMethods:
    def method1(self):
        pass
    
    def method2(self):
        pass

def BlankLineInFunction(self):
    return 7
    
    pass

#StartTest-blank_lines_in_for_loop
for i in range(2):
    pass
    
    pass
#EndTest

#StartTest-blank_line_in_try_catch
try:
    1
    
except:
    2
#EndTest

#StartTest-blank_line_in_try_catch_else
try:
    1
    
except:
    2
    
else:
    3
#EndTest

#StartTest-blank_trailing_line
def foo():
    return 1

#EndTest

def tabs():
    return 1



"""Non-keyboard events used in bpython curtsies REPL"""

import time
from collections.abc import Sequence

import curtsies.events


class ReloadEvent(curtsies.events.Event):
    """Request to rerun REPL session ASAP because imported modules changed"""

    def __init__(self, files_modified: Sequence[str] = ("?",)) -> None:
        self.files_modified = files_modified

    def __repr__(self) -> str:
        return "<ReloadEvent from {}>".format(" & ".join(self.files_modified))


class RefreshRequestEvent(curtsies.events.Event):
    """Request to refresh REPL display ASAP"""

    def __repr__(self) -> str:
        return "<RefreshRequestEvent for now>"


class ScheduledRefreshRequestEvent(curtsies.events.ScheduledEvent):
    """Request to refresh the REPL display at some point in the future

    Used to schedule the disappearance of status bar message that only shows
    for a few seconds"""

    def __init__(self, when: float) -> None:
        super().__init__(when)

    def __repr__(self) -> str:
        return "<RefreshRequestEvent for {} seconds from now>".format(
            self.when - time.time()
        )


class RunStartupFileEvent(curtsies.events.Event):
    """Request to run the startup file."""


class UndoEvent(curtsies.events.Event):
    """Request to undo."""

    def __init__(self, n: int = 1) -> None:
        self.n = n



import pydoc
import sys

from .pager import page

# Ugly monkeypatching
pydoc.pager = page  # type: ignore


class _Helper:
    def __init__(self) -> None:
        if hasattr(pydoc.Helper, "output"):
            # See issue #228
            self.helper = pydoc.Helper(sys.stdin, None)
        else:
            self.helper = pydoc.Helper(sys.stdin, sys.stdout)

    def __repr__(self) -> str:
        return (
            "Type help() for interactive help, "
            "or help(object) for help about object."
        )

    def __call__(self, *args, **kwargs) -> None:
        self.helper(*args, **kwargs)


_help = _Helper()


# vim: sw=4 ts=4 sts=4 ai et



import sys
from codeop import CommandCompiler
from typing import Any
from collections.abc import Iterable

from pygments.token import Generic, Token, Keyword, Name, Comment, String
from pygments.token import Error, Literal, Number, Operator, Punctuation
from pygments.token import Whitespace, _TokenType
from pygments.formatter import Formatter
from pygments.lexers import get_lexer_by_name
from curtsies.formatstring import FmtStr

from ..curtsiesfrontend.parse import parse
from ..repl import Interpreter as ReplInterpreter


default_colors = {
    Generic.Error: "R",
    Keyword: "d",
    Name: "c",
    Name.Builtin: "g",
    Comment: "b",
    String: "m",
    Error: "r",
    Literal: "d",
    Number: "M",
    Number.Integer: "d",
    Operator: "d",
    Punctuation: "d",
    Token: "d",
    Whitespace: "d",
    Token.Punctuation.Parenthesis: "R",
    Name.Function: "d",
    Name.Class: "d",
}


class BPythonFormatter(Formatter):
    """This is subclassed from the custom formatter for bpython.  Its format()
    method receives the tokensource and outfile params passed to it from the
    Pygments highlight() method and slops them into the appropriate format
    string as defined above, then writes to the outfile object the final
    formatted string. This does not write real strings. It writes format string
    (FmtStr) objects.

    See the Pygments source for more info; it's pretty
    straightforward."""

    def __init__(
        self,
        color_scheme: dict[_TokenType, str],
        **options: str | bool | None,
    ) -> None:
        self.f_strings = {k: f"\x01{v}" for k, v in color_scheme.items()}
        # FIXME: mypy currently fails to handle this properly
        super().__init__(**options)  # type: ignore

    def format(self, tokensource, outfile):
        o = ""

        for token, text in tokensource:
            while token not in self.f_strings:
                token = token.parent
            o += f"{self.f_strings[token]}\x03{text}\x04"
        outfile.write(parse(o.rstrip()))


class Interp(ReplInterpreter):
    def __init__(
        self,
        locals: dict[str, Any] | None = None,
    ) -> None:
        """Constructor.

        We include an argument for the outfile to pass to the formatter for it
        to write to.
        """
        super().__init__(locals)

        # typically changed after being instantiated
        # but used when interpreter used corresponding REPL
        def write(err_line: str | FmtStr) -> None:
            """Default stderr handler for tracebacks

            Accepts FmtStrs so interpreters can output them"""
            sys.stderr.write(str(err_line))

        self.write = write  # type: ignore
        self.outfile = self

    def writetb(self, lines: Iterable[str]) -> None:
        tbtext = "".join(lines)
        lexer = get_lexer_by_name("pytb")
        self.format(tbtext, lexer)
        # TODO for tracebacks get_lexer_by_name("pytb", stripall=True)

    def format(self, tbtext: str, lexer: Any) -> None:
        # FIXME: lexer should be "Lexer"
        traceback_informative_formatter = BPythonFormatter(default_colors)
        traceback_code_formatter = BPythonFormatter({Token: "d"})

        no_format_mode = False
        cur_line = []
        for token, text in lexer.get_tokens(tbtext):
            if text.endswith("\n"):
                cur_line.append((token, text))
                if no_format_mode:
                    traceback_code_formatter.format(cur_line, self.outfile)
                    no_format_mode = False
                else:
                    traceback_informative_formatter.format(
                        cur_line, self.outfile
                    )
                cur_line = []
            elif text == "    " and len(cur_line) == 0:
                no_format_mode = True
                cur_line.append((token, text))
            else:
                cur_line.append((token, text))
        assert cur_line == [], cur_line


def code_finished_will_parse(
    s: str, compiler: CommandCompiler
) -> tuple[bool, bool]:
    """Returns a tuple of whether the buffer could be complete and whether it
    will parse

    True, True means code block is finished and no predicted parse error
    True, False means code block is finished because a parse error is predicted
    False, True means code block is unfinished
    False, False isn't possible - an predicted error makes code block done"""
    try:
        return bool(compiler(s)), True
    except (ValueError, SyntaxError, OverflowError):
        return True, False



# The MIT License
#
# Copyright (c) 2015-2021 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import IO, Literal
from types import TracebackType


class BaseLock:
    """Base class for file locking"""

    def __init__(self) -> None:
        self.locked = False

    def acquire(self) -> None:
        pass

    def release(self) -> None:
        pass

    def __enter__(self) -> "BaseLock":
        self.acquire()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        if self.locked:
            self.release()
        return False

    def __del__(self) -> None:
        if self.locked:
            self.release()


try:
    import fcntl
    import errno

    class UnixFileLock(BaseLock):
        """Simple file locking for Unix using fcntl"""

        def __init__(self, fileobj, mode: int = 0) -> None:
            super().__init__()
            self.fileobj = fileobj
            self.mode = mode | fcntl.LOCK_EX

        def acquire(self) -> None:
            try:
                fcntl.flock(self.fileobj, self.mode)
                self.locked = True
            except OSError as e:
                if e.errno != errno.ENOLCK:
                    raise e

        def release(self) -> None:
            self.locked = False
            fcntl.flock(self.fileobj, fcntl.LOCK_UN)

    has_fcntl = True
except ImportError:
    has_fcntl = False


try:
    import msvcrt
    import os

    class WindowsFileLock(BaseLock):
        """Simple file locking for Windows using msvcrt"""

        def __init__(self, filename: str) -> None:
            super().__init__()
            self.filename = f"{filename}.lock"
            self.fileobj = -1

        def acquire(self) -> None:
            # create a lock file and lock it
            self.fileobj = os.open(
                self.filename, os.O_RDWR | os.O_CREAT | os.O_TRUNC
            )
            msvcrt.locking(self.fileobj, msvcrt.LK_NBLCK, 1)

            self.locked = True

        def release(self) -> None:
            self.locked = False

            # unlock lock file and remove it
            msvcrt.locking(self.fileobj, msvcrt.LK_UNLCK, 1)
            os.close(self.fileobj)
            self.fileobj = -1

            try:
                os.remove(self.filename)
            except OSError:
                pass

    has_msvcrt = True
except ImportError:
    has_msvcrt = False


def FileLock(
    fileobj: IO, mode: int = 0, filename: str | None = None
) -> BaseLock:
    if has_fcntl:
        return UnixFileLock(fileobj, mode)
    elif has_msvcrt and filename is not None:
        return WindowsFileLock(filename)
    return BaseLock()


# vim: sw=4 ts=4 sts=4 ai et



# The MIT License
#
# Copyright (c) 2015-2021 Sebastian Ramacher
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
from collections.abc import Iterator
from functools import cached_property
from re import Pattern, Match


class LazyReCompile:
    """Compile regular expressions on first use

    This class allows one to store regular expressions and compiles them on
    first use."""

    def __init__(self, regex: str, flags: int = 0) -> None:
        self.regex = regex
        self.flags = flags

    @cached_property
    def compiled(self) -> Pattern[str]:
        return re.compile(self.regex, self.flags)

    def finditer(self, *args, **kwargs) -> Iterator[Match[str]]:
        return self.compiled.finditer(*args, **kwargs)

    def search(self, *args, **kwargs) -> Match[str] | None:
        return self.compiled.search(*args, **kwargs)

    def match(self, *args, **kwargs) -> Match[str] | None:
        return self.compiled.match(*args, **kwargs)

    def sub(self, *args, **kwargs) -> str:
        return self.compiled.sub(*args, **kwargs)
